# -*- coding: utf-8 -*-

"""
Main file to define the CR labeling pipelien for LCO imaging data
CY Xu (cxu@ucsb.edu)
"""

import time
import numpy as np

from skimage import morphology
from skimage.segmentation import flood_fill
from scipy import ndimage

from cosmic_conn.cr_pipeline.utils_img import (
    sigma_clip,
    subtract_sky,
    erase_boundary_np,
)
from cosmic_conn.cr_pipeline.utils_img import sep_source_mask, center_crop_npy
from cosmic_conn.cr_pipeline.utils_io import (
    hdul_to_array,
    save_fits_with_CR,
    save_as_png,
    save_array_png,
)
from cosmic_conn.cr_pipeline.utils_io import remove_outlier_frm
from cosmic_conn.cr_pipeline.banzai_stats import median_absolute_deviation


class CR_reduction:
    def __init__(self, hdul=None, opt=None):
        self.hdul = hdul
        self.opt = opt
        self.reject_dir = opt.root + "/rejected_frms"
        self.pre_process()

    def pre_process(self):
        header = self.hdul["SCI"].header

        self.fname = header["filename"]
        self.exp_time = header["exptime"]
        self.frmtotal = header["frmtotal"]
        self.min_cr_size = self.opt.min_cr_size
        rdnoise = header["rdnoise"]

        # read hdul as ndarrays of shape N,H,W
        self.frames_banzai, self.valid_mask, gains, self.filenames = hdul_to_array(
            self.hdul
        )

        # pre-processing, a separate copy for mask generation
        frames = sigma_clip(self.frames_banzai.copy(),
                            sigma=3.0, clip_source=False)

        frames_median_subtracted = subtract_sky(frames.copy())
        self.source_masks = sep_source_mask(
            frames_median_subtracted, self.valid_mask)

        # re-calculate gain for each 0m4 exposure with source mask
        if "0m4" in self.fname and self.source_masks is not None:
            # sovle for the gain for each image
            self.new_gains = self.sovle_for_gain(
                frames, rdnoise, self.source_masks)

            # for 0m4 data, update image and read_noise with new gain
            self.frames = frames * self.new_gains
            self.rdnoises = rdnoise * self.new_gains

            with open(self.opt.log_file, "a") as log:
                msg = f"{self.fname}, new gains {list(self.new_gains.ravel())}\n\n"
                log.write(msg)
                print(msg)
        else:
            self.frames = frames
            self.rdnoises = rdnoise
            self.new_gains = np.array(gains).reshape(frames.shape[0], 1, 1)

        # remove no_data pixels, potential bad boundary CCD pixels (100px),
        # and potential bad SEP analysis near the boundary, included in ignore mask
        self.ignore_boundary = 100

        self.frames *= self.valid_mask
        self.frames = erase_boundary_np(
            self.frames, boundary_width=self.ignore_boundary
        )

    def generate_CR_masks(self):
        """
        This is the main CR detection loop
            estimate_reject_refine() is called one or multiple times for CR detection
            detect CR in the sequence of frames, reject outlier pixels and frames

        if estimate_reject_refine() returns a string, it's a error message during detection
        print the error message and abort this sequence
        """
        tic = time.perf_counter()
        high_thres = self.opt.snr_thres
        low_thres = self.opt.snr_thres_low

        # return error message if source mask not valid
        if self.source_masks is None:
            return "rejected_SEP_sources_over_3k"

        # initial estimation to reject outlier frame
        cr_frac = self.estimate_reject_refine(high_thres, low_thres)

        if isinstance(cr_frac, str):
            self.output_png_preview(reason=cr_frac)
            return cr_frac

        # if an outlier frame is rejected, detect again with the new sequence because the
        # CR frac between frames might change, until all frames considered well aligned
        while len(self.reject_idx) > 0:
            cr_frac = self.estimate_reject_refine(high_thres, low_thres)
            if isinstance(cr_frac, str):
                return cr_frac

        # input frames exposure time capped at 100s, no frame should have CR less than 1e-5
        if cr_frac < 1e-5:
            self.output_png_preview(reason="rejected_low_cr_frac")
            return f"CR frac too low {cr_frac}, exposure time {self.exp_time}"

        # all rejection passed, print stats of the useable frames
        with open(self.opt.log_file, "a") as log:
            msg = (
                f"{self.fname}, exp_time {self.exp_time}\n"
                f"CR count {[np.sum(m) for m in self.cr_mask]}, CR fraction {round(cr_frac, 6)}\n\n"
            )
            log.write(msg)
            print(msg)

        debug_frames = self.mask_high, self.mask_low, self.snr_deviation

        # if everythings checks out, output new fits with masks
        save_fits_with_CR(
            self.opt,
            self.hdul,
            self.new_gains,
            self.cr_mask,
            self.ignore_mask,
            self.filenames,
            debug_frames,
        )

        # replace CR pixel with median frame and output png for quick preview
        self.clean_frames = np.where(
            self.cr_mask > 0, self.median_frame, self.frames)

        # output png preview or numpy as training set (no longer using numpy)
        if not self.opt.no_png:
            self.output_png_preview(self.opt.comment)

    def estimate_reject_refine(self, high_thres, low_thres, reject_frm=True):
        """
        The main CR detection happens here:
            1. estimate two masks based on an initial SNR threshold and
               a lower threshold (for peripheral CR edges)

            2. refine CR to incldue soft edges via dilation/flood filling

            3. generate the ignore mask, definition:
                OK (0)
                boundary (1)
                no data (2)
                sources (4)
                hot pixels (8)

            4. multiply CR mask with ignore mask again

            5. reject frame outliers with excessive CR in a sequence
               often false positives caused by satellite trails or mis-aligned frames 
        """

        # 1.
        mask_low, mask_high = self.estimate_two_masks(
            self.frames, self.rdnoises, high_thres, low_thres
        )
        # from here on, self.frames are sky subtracted

        # 2.
        self.cr_mask, single_pixels = self.maximize_cr_edges(
            mask_low, mask_high, self.min_cr_size
        )

        # return error message
        if isinstance(self.cr_mask, str):
            return self.cr_mask

        # 3. ignore mask include sources, single hot pixels, and empty boundary
        self.ignore_mask = erase_boundary_np(
            np.zeros_like(self.frames, dtype="uint8"),
            boundary_width=self.ignore_boundary,
            replace_value=1,
        )
        self.ignore_mask += (1 - self.valid_mask) * 2
        self.ignore_mask += self.source_masks * 4
        self.ignore_mask += single_pixels * 8

        # 4.
        self.cr_mask *= (self.ignore_mask == 0).astype("uint8")

        # 5.
        if reject_frm:
            self.reject_idx = self.reject_outlier_frm()
        else:
            self.reject_idx = []

        # return error message
        if self.frmtotal < 3:
            return "rejected_less_than_3_frames"

        # calculate CR fraction/ratio, used for adaptive SNR threshold
        total_pixels = self.cr_mask.size
        cr_frac = np.sum(self.cr_mask) / total_pixels

        return cr_frac

    def compute_SNR(self, frames, read_noise):
        # assuming frames in an N_images x Ny x Nx array
        N_frames = frames.shape[0]

        bkg_noise = np.median(frames, axis=(1, 2), keepdims=True)

        # calculate uncertainties
        frame_uncertainty = np.sqrt(
            read_noise ** 2 + np.abs(frames) + bkg_noise)

        median_frame_uncertainty = np.median(frame_uncertainty, axis=0) / np.sqrt(
            N_frames
        )

        # calculate the total noise including our model (approximately for the median)
        total_uncertainty = np.sqrt(
            frame_uncertainty ** 2 + median_frame_uncertainty ** 2
        )

        # subtract sky and get correct median
        frames = subtract_sky(frames, remove_negative=False)
        self.median_frame = np.median(frames, axis=0)

        # The median acts as our model of what the frame should look like,
        # so we can look for deviations from that
        # without np.abs(frames), only consider positive deviations
        snr_deviation = (frames - self.median_frame) / total_uncertainty

        return snr_deviation

    def estimate_two_masks(self, frames, read_noise, high_thres, low_thres):
        """
        noisy_stack: ndarray
            a stack of noisy frames of shape (n, h, w)

        Rerturn:
            mask_low, retrieved from a lower threshold, more pixel detected but more false positive as well
            mask_high, retrieved from a higher threshold to reject false positive
        """

        snr_deviation = self.compute_SNR(frames, read_noise)

        mask_high = (snr_deviation > high_thres).astype("uint8")

        # lower threshold to get the soft edges of CR
        mask_low = (snr_deviation > low_thres).astype("uint8")

        # for debugging
        self.snr_deviation = snr_deviation
        self.mask_high, self.mask_low = mask_high, mask_low

        return mask_low, mask_high

    def sovle_for_gain(self, frames, rdnoise, source_mask):
        # if no vlaid source masks, return 1 for now, this sequence will be skipped
        if source_mask is None:
            return np.array(np.ones((frames.shape[0], 1, 1)))

        # ignore sources and hot pixels
        bkg_imgs = frames * (1 - source_mask)
        new_gains = []

        for i in range(frames.shape[0]):
            # crop 100px boundary
            bkg = center_crop_npy(bkg_imgs[i], crop_ratio=0.9)

            mad = median_absolute_deviation(bkg)
            median = np.median(bkg)

            # solve for (mad * 1.48)**2 - gain * median - gain**2 * read_noise ** 2 = 0
            coeff = [-1 * rdnoise ** 2, -1 * median, (mad * 1.48) ** 2]
            gain = np.roots(coeff)[1]
            new_gains.append(round(gain, 6))

        return np.array(new_gains).reshape(-1, 1, 1)

    def maximize_cr_edges(self, mask_low, mask_high, min_cr_size):
        shape = mask_high.shape

        # reject CR candidates over-lapping with sources
        mask_high = mask_high * (1 - self.source_masks)

        if self.opt.flood_fill:
            # Flood fill captures less false positive isolated pixels
            # but it also misses small true positive CRs
            mask_common = np.logical_and(mask_high, mask_low)

            for i in range(shape[0]):
                # create labels in mask_high
                imglab, total = morphology.label(
                    mask_common[i].astype("int"),
                    background=0,
                    return_num=True,
                    connectivity=2,
                )

                # any pixle insdie idx can be the starting point for flood fill
                for j in range(1, total + 1, 1):
                    index = np.where(imglab == j)
                    idx = (index[0][0], index[1][0])
                    mask_low[i] = flood_fill(mask_low[i], idx, 2)

            cr_mask = mask_low == 2

        else:
            # dilation expands CR edges by N pixels, more conservative than flood fill
            # isolated false positive pixels are included in the ignroe mask
            struct = ndimage.generate_binary_structure(2, 2)
            mask_dilation = np.zeros_like(mask_high)

            for i in range(shape[0]):
                mask_dilation[i] = ndimage.binary_dilation(
                    mask_high[i], structure=struct, iterations=self.opt.dilation
                )

            cr_mask = np.logical_and(mask_dilation, mask_low)

        # second pass rejection in case source included after dialtion
        cr_mask = cr_mask * (1 - self.source_masks)

        # save before removing single pixels to add them to ignore mask
        cr_mask = cr_mask.astype("bool")
        cr_mask_single_pixels = cr_mask.copy()

        # remove flagged CR smaller than certain pixels from final mask
        if min_cr_size > 1:
            for i in range(shape[0]):
                # create labels in segmented image
                cr_mask[i] = morphology.remove_small_objects(
                    cr_mask[i], min_size=min_cr_size, connectivity=2
                )

        # add single pixels and source mask to make the ignore mask
        single_pixels = np.logical_xor(
            cr_mask_single_pixels, cr_mask).astype("uint8")
        cr_mask = cr_mask.astype("uint8")

        return cr_mask, single_pixels

    def reject_outlier_frm(self):
        shape = self.cr_mask.shape

        # cr counts in each frame
        cr_counts = np.sum(self.cr_mask, axis=(1, 2))

        total_cr = 1 if np.sum(cr_counts) == 0 else np.sum(cr_counts)
        frame_cr_contribution = cr_counts / total_cr

        # if the # of CR from one frames is more than any of the two conditions
        #   1. twice the average,
        #   2. two standard deviation of avergae contribution (outlier)
        # reject this frame, as it might have satellite trail, mis-aligned, or blurry

        mean = np.mean(frame_cr_contribution)
        reject_thres = min(2.0 * mean, mean + 2.0 *
                           np.std(frame_cr_contribution))

        reject_idx = np.where(frame_cr_contribution > reject_thres)[0].tolist()

        # save rejected frames for reference
        if len(reject_idx) > 0:

            for i in range(shape[0]):
                name = "noisy_rejected" if i in reject_idx else "noisy_accepted"
                save_as_png(
                    self.frames[i],
                    self.reject_dir,
                    self.filenames[i],
                    name,
                    zscale=True,
                )

            with open(self.opt.log_file, "a") as log:
                msg = (
                    f"{self.fname} rejected, reject_thres {reject_thres}\n"
                    f"{frame_cr_contribution}\n\n"
                )
                log.write(msg)
                print(msg)

        # remove rejected frames from frame stack and fits
        if len(reject_idx) > 0:
            self.frames = remove_outlier_frm(self.frames, reject_idx)
            self.frmtotal = self.frames.shape[0]

            self.frames_banzai = remove_outlier_frm(
                self.frames_banzai, reject_idx)

            self.cr_mask = remove_outlier_frm(self.cr_mask, reject_idx)

            self.filenames = remove_outlier_frm(
                np.array(self.filenames), reject_idx
            ).tolist()

            self.hdul = remove_outlier_frm(
                self.hdul, reject_idx, update_hdul=True)

        return reject_idx

    def output_png_preview(self, reason):
        if "reject" in reason:
            save_array_png(
                self.frames_banzai,
                None,
                None,
                None,
                self.opt.root,
                self.filenames,
                save_np=False,
                save_png=True,
                header=self.hdul[1].header,
                comment=reason,
            )
        else:
            save_array_png(
                self.frames,
                self.cr_mask,
                self.clean_frames,
                self.ignore_mask,
                self.opt.root,
                self.filenames,
                save_np=False,
                save_png=True,
                header=self.hdul[1].header,
                comment=reason,
            )



"""Top-level package for cosmic_conn."""

__author__ = """Chengyuan Xu, Curtis McCully, Boning Dong, D. Andrew Howell, and Pradeep Sen"""
__email__ = "cxu@ucsb.edu"
__version__ = "0.2.5"

from cosmic_conn.inference_cr import init_model, detect_image, detect_FITS


"""
pytorch tensor based utilities.
CY Xu (cxu@ucsb.edu)
"""
import os
import math
import datetime
import random
import psutil
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from cosmic_conn.cr_pipeline.banzai_stats import robust_standard_deviation


def tensor2np(gpu_tensor):
    # convert torch tensor to numpy array
    if isinstance(gpu_tensor, torch.Tensor):
        return gpu_tensor.detach().cpu().numpy()
    else:
        return gpu_tensor


def remove_nan(image):
    # replace NaN with 0.0 if exist
    if np.sum(np.isnan(image)) > 0:
        image = np.nan_to_num(image, copy=False, nan=0.0)
    return image


def memory_check(device):
    # check if the machien/GPU has sufficient memory for
    # full image detection without movign to swap
    # otherwise the image is sliced into stamps

    CPU_THRESHOLD = 16 * (1024**3)  # 16 GB free memory
    GPU_THRESHOLD = 8 * (1024**3)  # 8 GB free memory

    if str(device) == 'cpu':
        # on CPU, available memory
        available_memory = psutil.virtual_memory()[1]
        full_image_detection = available_memory > CPU_THRESHOLD
    else:
        # GPU available memory
        t = torch.cuda.get_device_properties(device).total_memory
        # r = torch.cuda.memory_reserved(device)
        # a = torch.cuda.memory_allocated(device)
        # available_memory = r-a  # free inside reserved
        full_image_detection = t > GPU_THRESHOLD

    if not full_image_detection:
        print(f"...available memory not sufficient for whole image detection.")
        print(f"...image will be sliced into stamps.")

    return full_image_detection


# https://www.kaggle.com/bigironsphere/loss-function-library-keras-pytorch
def DiceLoss(prediction, targets, smooth=1):

    # comment out if your model contains a sigmoid or equivalent activation layer
    # inputs = F.sigmoid(inputs)

    # prediction = median_weight * prediction

    # flatten label and prediction tensors
    prediction = prediction.contiguous().view(-1)
    targets = targets.contiguous().view(-1)

    intersection = (prediction * targets).sum()
    dice = (2.0 * intersection + smooth) / \
        (prediction.sum() + targets.sum() + smooth)

    return 1 - dice


# From https://discuss.pytorch.org/t/is-there-anyway-to-do-gaussian-filtering-for-an-image-2d-3d-in-pytorch/12351/3
def get_gaussian_kernel(kernel_size=3, sigma=2, channels=3):
    # Create a x, y coordinate grid of shape (kernel_size, kernel_size, 2)
    x_coord = torch.arange(kernel_size)
    x_grid = x_coord.repeat(kernel_size).view(kernel_size, kernel_size)
    y_grid = x_grid.t()
    xy_grid = torch.stack([x_grid, y_grid], dim=-1).float()

    mean = (kernel_size - 1) / 2.0
    variance = sigma ** 2.0

    # Calculate the 2-dimensional gaussian kernel which is
    # the product of two gaussian distributions for two different
    # variables (in this case called x and y)
    gaussian_kernel = (1.0 / (2.0 * math.pi * variance)) * torch.exp(
        -torch.sum((xy_grid - mean) ** 2.0, dim=-1) / (2 * variance)
    )

    # Make sure sum of values in gaussian kernel equals 1.
    gaussian_kernel = gaussian_kernel / torch.sum(gaussian_kernel)

    # Reshape to 2d depthwise convolutional weight
    gaussian_kernel = gaussian_kernel.view(1, 1, kernel_size, kernel_size)
    gaussian_kernel = gaussian_kernel.repeat(channels, 1, 1, 1)

    gaussian_filter = nn.Conv2d(
        in_channels=channels,
        out_channels=channels,
        kernel_size=kernel_size,
        padding=kernel_size // 2,
        groups=channels,
        bias=False,
    )

    gaussian_filter.weight.data = gaussian_kernel
    gaussian_filter.weight.requires_grad = False

    if torch.cuda.is_available():
        gaussian_filter.to(torch.device("cuda"))

    return gaussian_filter


# permanent
gaussian_filter = get_gaussian_kernel(kernel_size=5, sigma=2, channels=1)

# https://discuss.pytorch.org/t/solved-class-weight-for-bceloss/3114/2


def median_weighted_bce(p, y, median_stack=None, imbalance_alpha=1.0):
    eps = 1e-7

    # prevent gradient diminishing or explosion
    p = torch.clamp(p, eps, (1 - eps))

    if median_stack is None:
        loss = y * torch.log(p) + \
            (1 - y) * torch.log(1 - p)
    else:
        assert median_stack.shape == p.shape

        median_weighted_mask = torch.zeros_like(median_stack)

        if imbalance_alpha < 1.0:
            with torch.no_grad():

                # batch size for loop
                for i in range(median_stack.shape[0]):

                    median_curr = median_stack[i: i + 1]

                    robust_std = robust_standard_deviation(
                        tensor2np(median_curr.squeeze())
                    ) + eps

                    median_curr -= torch.median(median_curr)

                    floor = max(0, robust_std)
                    ceil = min(median_curr.max().item(), 5 * robust_std)

                    mask_curr = torch.clamp(median_curr, floor, ceil)
                    mask_curr -= floor

                    mask_curr = gaussian_filter(mask_curr)

                    # avoid zero division
                    if mask_curr.max() > 0:
                        mask_curr /= mask_curr.max()

                    mask_curr = torch.clamp(mask_curr, imbalance_alpha, 1.0)

                    # populatet the processed mask to a stack
                    median_weighted_mask[i] = mask_curr

        else:
            median_weighted_mask = 1.0

        p = p.contiguous()
        y = y.contiguous()

        # median weighted BCE
        loss = (y * torch.log(p)) + \
            median_weighted_mask * ((1 - y) * torch.log(1 - p))

    return torch.neg(torch.mean(loss))


# https://www.kaggle.com/c/tgs-salt-identification-challenge/discussion/65938
class FocalLoss(nn.Module):
    def __init__(self, alpha=1, gamma=2, logits=False):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.logits = logits

    def forward(self, inputs, targets):
        if self.logits:
            BCE_loss = F.binary_cross_entropy_with_logits(
                inputs, targets, reduction="none"
            )
        else:
            BCE_loss = F.binary_cross_entropy(
                inputs, targets, reduction="none")
        pt = torch.exp(-BCE_loss)
        F_loss = self.alpha * (1 - pt) ** self.gamma * BCE_loss
        return torch.mean(F_loss)


def replace_boundary(array, boundary, replace_value=0):
    array[0:boundary, :] = replace_value
    array[-boundary:, :] = replace_value
    array[:, 0:boundary] = replace_value
    array[:, -boundary:] = replace_value
    return array


def erase_boundary_tensor(tensor, boundary_width=0):
    in_shape = tensor.shape

    # replace boundary with 0 to avoid occasional issues on CCD boundary
    if boundary_width == 0:
        return tensor

    if boundary_width > 0:
        reshaped = tensor.view(-1, in_shape[-2], in_shape[-1])

        for i in range(reshaped.shape[0]):
            reshaped[i] = replace_boundary(reshaped[i], boundary_width)

    tensor = reshaped.view(in_shape)

    return tensor


def erase_boundary_np(ndarray, boundary_width=0, replace_value=0):
    # replace boundary with 0 to avoid occasional issues on CCD boundary
    if boundary_width == 0:
        return ndarray

    if boundary_width > 0:
        copy = ndarray.copy()

        if len(copy.shape) == 2:
            copy = replace_boundary(copy, boundary_width, replace_value)
            return copy

        elif len(copy.shape) == 3:
            for i in range(copy.shape[0]):
                copy[i] = replace_boundary(
                    copy[i], boundary_width, replace_value)
            return copy


def clean_large(image, model, patch=1024, overlap=0, ret_numpy=False):
    """
        modified from deepCR implementation
        given input image
        return cosmic ray mask and (optionally) clean image
        mask could be binary or probabilistic
    :param img0: (np.ndarray) 2D input image
    :return: mask or binary mask; or None if internal call
    """
    if not isinstance(image, torch.Tensor):
        image = torch.tensor(image).float()

    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    image = image.to(device)

    shape = image.shape
    stride = patch - 2 * overlap
    hh = math.ceil(shape[0] / stride)
    ww = math.ceil(shape[1] / stride)
    mask = torch.zeros_like(image)

    for i in range(hh):
        for j in range(ww):
            # overlapping crop at stride
            h_start = i * stride
            w_start = j * stride
            h_stop = min(h_start + patch, shape[0])
            w_stop = min(w_start + patch, shape[1])

            crop = image[h_start:h_stop, w_start:w_stop]
            crop_h, crop_w = crop.shape

            # sky subtraction
            crop -= crop.median()

            if crop_h < patch or crop_w < patch:
                crop = modulus_boundary_crop(crop)

            if 0 in crop.shape:
                continue

            # expand dimension to fit the model
            pdt = model(dim_expand(crop))
            pdt = pdt.squeeze()

            h_pad = min(1, i) * overlap
            w_pad = min(1, j) * overlap

            h_start_ = h_pad + h_start
            w_start_ = w_pad + w_start

            effective_pdt = pdt[h_pad:, w_pad:]

            h_stop_ = h_start_ + effective_pdt.shape[0]
            w_stop_ = w_start_ + effective_pdt.shape[1]

            mask[h_start_:h_stop_, w_start_:w_stop_] = effective_pdt

    if ret_numpy:
        mask = tensor2np(mask)

    return mask


def center_crop(tensor, crop_size=0):
    # center cropping for efficiency
    if crop_size == 0:
        return tensor

    h, w = tensor.shape

    if h < crop_size or w < crop_size:
        return tensor

    h_offset = (h - crop_size) // 2
    w_offset = (w - crop_size) // 2
    h_stop = min(h_offset + crop_size, h)
    w_stop = min(w_offset + crop_size, w)

    return tensor[h_offset:h_stop, w_offset:w_stop]


def center_crop_npy(array_in, crop_size=0):
    # center crop a quare img from 2D array
    if crop_size == 0:
        return array_in
    else:
        h, w = array_in.shape
        assert h > crop_size and w > crop_size, ValueError(
            f"array smaller than crop size"
        )

        cropped = np.zeros((crop_size, crop_size))

        left, top = int(w / 2 - crop_size / 2), int(h / 2 - crop_size / 2)
        cropped = array_in[top: top + crop_size, left: left + crop_size]

        return cropped


def modulus_boundary_crop(array, modulo=16):
    """
    crop a largest possible area of N modulo
    to simply the problem, start from top-left corner to the max valid area
    expect input as a 2D array
    """
    h, w = array.shape[0], array.shape[1]
    new_h = h - (h % modulo)
    new_w = w - (w % modulo)

    if new_h == h and new_w == w:
        return array

    new_array = array[:new_h, :new_w]

    return new_array


def dim_expand(tensor):
    if len(tensor.shape) == 4:
        return tensor

    elif len(tensor.shape) == 2:
        return tensor.unsqueeze(dim=0).unsqueeze(dim=0)

    elif len(tensor.shape) == 3:
        return tensor.unsqueeze(dim=0)

    else:
        raise ValueError(
            f"tensor shape {tensor.shape} is not valid, \
            array shape [n, 1, h, w] is expected."
        )


# PyTroch version
# https://www.kaggle.com/iezepov/fast-iou-scoring-metric-in-pytorch-and-numpy
SMOOTH = 1e-6


def iou_pytorch(outputs: torch.Tensor, labels: torch.Tensor):
    # You can comment out this line if you are passing tensors of equal shape
    # But if you are passing output from UNet or something it will most probably
    # be with the BATCH x 1 x H x W shape
    outputs = outputs.squeeze(1)  # BATCH x 1 x H x W => BATCH x H x W

    intersection = (
        (outputs & labels).float().sum((1, 2))
    )  # Will be zero if Truth=0 or Prediction=0
    union = (outputs | labels).float().sum(
        (1, 2))  # Will be zzero if both are 0

    iou = (intersection + SMOOTH) / (
        union + SMOOTH
    )  # We smooth our devision to avoid 0/0

    thresholded = (
        torch.clamp(20 * (iou - 0.5), 0, 10).ceil() / 10
    )  # This is equal to comparing with thresolds

    return thresholded  # Or thresholded


def _worker_init_fn(worker_id: int) -> None:
    """Sets a unique but deterministic random seed for background workers.

    Only sets the seed for NumPy because PyTorch and Python's own RNGs
    take care of reseeding on their own.
    See https://github.com/numpy/numpy/issues/9650."""
    # Modulo 2**32 because np.random.seed() only accepts values up to 2**32 - 1
    initial_seed = torch.initial_seed() % 2 ** 32
    worker_seed = initial_seed + worker_id
    np.random.seed(worker_seed)


def extract_random_frame(stacks):
    # randomly pick one frame from a sequence of three frames
    batch = stacks.shape[0]
    frames = stacks.shape[1]
    idx = random.randint(0, frames - 1)

    frames = [stacks[i, idx] for i in range(batch)]
    frames = torch.stack(frames).unsqueeze(dim=1)

    return frames


def subtract_sky(tensor, remove_negative=False):
    in_shape = tensor.shape

    tensor = tensor.view(-1, in_shape[-2], in_shape[-1])

    for i in range(tensor.shape[0]):
        tensor[i] -= tensor[i].median()

    if remove_negative:
        tensor = torch.clamp(tensor, 0.0, tensor.max())

    tensor = tensor.view(in_shape)

    return tensor


def tensor2uint8(gpu_tensor, dtype="uint8", method="clipping"):
    dims = len(gpu_tensor.shape)

    if dims == 3:
        array = tensor2np(gpu_tensor)
    elif dims == 2:
        array = tensor2np(gpu_tensor.unsqueeze(0))

    if method == "clipping":
        img = np.clip(array, 0, 255)
    elif method == "normalize":
        img = array - array.min()
        img = img / img.max()
        img *= 255

    return img.astype(dtype)


def remove_ext(filename):
    if filename.endswith(".fz"):
        return filename[:-8]
    elif filename.endswith(".fitz"):
        return filename[:-5]
    else:
        return os.path.splitext(filename)[0]


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def mkCheckpointDir(suffix, continue_train):
    if continue_train:
        directory = continue_train
        path = f"./checkpoints/{continue_train}"

    else:
        now = datetime.datetime.now()
        timestamp = now.strftime("%Y_%m_%d_%H_%M")
        directory = str(timestamp) + "_" + suffix
        path = f"./checkpoints/{directory}"
        mkdir(path)

    return path, directory


EXTENSIONS = ["npz", "npy"]


def is_numpy_file(filename):
    filename = filename.lower()
    return any(filename.endswith(extension) for extension in EXTENSIONS)


def read_npy(directory):
    # read from multiple sub-dirs, output a sequential list of noisy images, masks and file names
    noisy_frms = []
    masks = []
    filenames = []

    for root, dirs, files in os.walk(directory):
        if "mask" in root and len(files) > 0:
            files.sort()
            for f in files:
                if f.endswith("mask.npy"):
                    filenames.append(f)
                    mask_path = os.path.join(root, f)
                    noisy_path = os.path.join(root, f.replace("mask", "noisy"))
                    try:
                        # if shape_n[0] < 2000 or shape_m[0] < 2000:
                        #     print(shape_m, shape_n)
                        #     print(f'{f} size smaller than sensor, file (not) removed')
                        # else:
                        noisy_frms.append(noisy_path)
                        masks.append(mask_path)
                    except:
                        print(
                            f"{f} is missing mask/noisy pair, file (not) removed")
                        # os.remove(path)
                        continue

    dataset_size = len(noisy_frms)
    assert dataset_size > 0, f"No valid training data found in {directory}"
    print(f"Training dataset size: {dataset_size}")

    return noisy_frms, masks, filenames


def gen_train_test_data(noisy_frms, masks, filenames, train_ratio=0.9):
    dataset_size = len(noisy_frms)
    train_size = int(dataset_size * train_ratio)

    train_idx = np.random.choice(
        dataset_size, train_size, replace=False).tolist()
    # complement of train_idx is the test idx
    test_idx = list(set(np.arange(dataset_size).tolist()) - set(train_idx))

    noisy_train = [noisy_frms[i] for i in train_idx]
    mask_train = [masks[i] for i in train_idx]
    file_train = [filenames[i] for i in train_idx]

    cropped_test = [noisy_frms[i] for i in test_idx]
    mask_test = [masks[i] for i in test_idx]
    file_test = [filenames[i] for i in test_idx]

    return noisy_train, mask_train, file_train, cropped_test, mask_test, file_test



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

# with open("HISTORY.rst") as history_file:
#     history = history_file.read()

requirements = [
    "torch>=1.6.0",
    "numpy",
    "astropy>=3.0",
    "reproject", 
    "sep",
    "scikit-image",
    "psutil",
    "pretty-errors",
    "tqdm",
    "requests",
]

extras_require = {    
    "webapp": ["Flask>=1.1.0", "Flask-APScheduler>=1.12.0"],
    "develop": ["tensorboard>=2.4.0", "scikit-learn>=0.24.0", "Flask>=1.1.0", "Flask-APScheduler>=1.12.0"],
}

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
    "pip",
    "bump2version",
    "wheel",
    "watchdog",
    "flake8",
    "tox",
    "coverage",
    "Sphinx",
    "twine",
]

setup(
    author="Chengyuan Xu, Curtis McCully, Boning Dong, D. Andrew Howell, and Pradeep Sen",
    author_email="cxu@ucsb.edu",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.7",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Environment :: GPU :: NVIDIA CUDA",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    description="Cosmic-CoNN: A Cosmic Ray Detection Deep Learning Framework, Dataset, and Toolkit",
    entry_points={
        "console_scripts": [
            "cosmic-conn=cosmic_conn.inference_cr:CLI_entry_point",
            "cosmic_conn=cosmic_conn.inference_cr:CLI_entry_point",
        ],
    },
    install_requires=requirements,
    extras_require=extras_require,
    dependency_links=[
        '-f https://download.pytorch.org/whl/torch_stable.html'],
    license="GNU General Public License v3",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords="cosmic_conn",
    name="cosmic_conn",
    packages=find_packages(include=[
        "cosmic_conn",
        "cosmic_conn.*",
        "cr_pipeline",
        "dl_framework",
        "evaluate",
        "web_app",
    ]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/cy-xu/cosmic-conn",
    version="0.2.5",
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""
Defines command line arguments parser for CR labeling pipeline
CY Xu (cxu@ucsb.edu)
"""

import argparse


def argument_parser():
    """
    options for CR mask generation, default values are recommended, example command:
    `python stack_clean.py data/sample_path --sigma 5 --min_cr_size 3 --verbose`
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data",
        default="./data",
        help="path to banzai fits, banzai fits much be placed in banzai_frms directory",
    )

    parser.add_argument(
        "--min_exptime",
        type=float,
        default=100.0,
        help="minimum exposure time used to generate the CR mask",
    )

    parser.add_argument(
        "--snr_thres", type=float, default=5.0, help="initial SNR threshold"
    )

    parser.add_argument(
        "--snr_thres_low", type=float, default=2.5, help="initial SNR threshold"
    )

    parser.add_argument(
        "--dilation",
        type=int,
        default=1,
        help="minimum continuum CR size in pixels, used to reject very small CR that are possibily false positive",
    )

    parser.add_argument(
        "--flood_fill",
        action="store_true",
        help="use flood fill algorithm to expand peripheral pixels",
    )

    parser.add_argument(
        "--aligned",
        action="store_true",
        help="frames already aligned and saved in aligned_fits directroy",
    )

    parser.add_argument(
        "--min_cr_size",
        type=int,
        default=3,
        help="minimum continuum CR size in pixels, used to reject very small CR that are possibily false positive",
    )

    parser.add_argument(
        "--comment", default=None, help="remarks saved at end of export filename"
    )

    parser.add_argument(
        "--no_png", action="store_true", help="do NOT export png if flagged"
    )

    parser.add_argument(
        "--linear_model",
        action="store_true",
        help="do NOT use the linear CR model, simply the initial Sigma (5 by default)",
    )

    parser.add_argument(
        "--scrappy", action="store_true", help="generate CR from astro_scrappy"
    )

    parser.add_argument(
        "--verbose", action="store_true", help="print more info for debug"
    )

    parser.add_argument(
        "--debug", action="store_true", help="print more info for debug"
    )

    parser.add_argument(
        "--nres", action="store_true", help="process NRES Spectrographs"
    )

    parser.add_argument(
        "--cpus", type=int, default=1, help="use multiple CPU cores for parallism"
    )

    return parser.parse_args()
